@@ -12,6 +12,7 @@ import signal
 import sys
 import threading
 import traceback
+from collections import namedtuple
 
 import py
 import pytest
@@ -31,6 +32,13 @@ Timeout mechanism to use.  'signal' uses SIGALRM if available,
 'thread' uses a timer thread.  The default is to use 'signal' and fall
 back to 'thread'.
 """.strip()
+FUNC_ONLY_DESC = """
+When set to True, defers the timeout evaluation to only the test
+function body, ignoring the time it takes when evaluating any fixtures
+used in the test.
+""".strip()
+
+Settings = namedtuple('Settings', ['timeout', 'method', 'func_only'])
 
 
 @pytest.hookimpl
@@ -40,21 +48,20 @@ def pytest_addoption(parser):
         'timeout', 'Interrupt test run and dump stacks of all threads after '
         'a test times out')
     group.addoption('--timeout',
-                    type='int',
+                    type=float,
                     help=TIMEOUT_DESC)
     group.addoption('--timeout_method',
-                    type='choice',
                     action='store',
                     choices=['signal', 'thread'],
-                    help='Depreacted, use --timeout-method')
+                    help='Deprecated, use --timeout-method')
     group.addoption('--timeout-method',
                     dest='timeout_method',
-                    type='choice',
                     action='store',
                     choices=['signal', 'thread'],
                     help=METHOD_DESC)
     parser.addini('timeout', TIMEOUT_DESC)
     parser.addini('timeout_method', METHOD_DESC)
+    parser.addini('timeout_func_only', FUNC_ONLY_DESC, type='bool')
 
 
 @pytest.hookimpl
@@ -62,17 +69,48 @@ def pytest_configure(config):
     # Register the marker so it shows up in --markers output.
     config.addinivalue_line(
         'markers',
-        'timeout(timeout, method=None): Set a timeout and timeout method on '
-        'just one test item.  The first argument, *timeout*, is the timeout '
-        'in seconds while the keyword, *method*, takes the same values as the '
-        '--timeout_method option.')
+        'timeout(timeout, method=None, func_only=False): Set a timeout, timeout '
+        'method and func_only evaluation on just one test item.  The first '
+        'argument, *timeout*, is the timeout in seconds while the keyword, '
+        '*method*, takes the same values as the --timeout_method option. The '
+        '*func_only* keyword, when set to True, defers the timeout evaluation '
+        'to only the test function body, ignoring the time it takes when '
+        'evaluating any fixrures used in the test.')
+
+    settings = get_env_settings(config)
+    config._env_timeout = settings.timeout
+    config._env_timeout_method = settings.method
+    config._env_timeout_func_only = settings.func_only
+
 
 
 @pytest.hookimpl(hookwrapper=True)
 def pytest_runtest_protocol(item):
-    timeout_setup(item)
+    func_only = get_func_only_setting(item)
+    if func_only is False:
+        timeout_setup(item)
     outcome = yield
-    timeout_teardown(item)
+    if func_only is False:
+        timeout_teardown(item)
+
+
+@pytest.hookimpl(hookwrapper=True)
+def pytest_runtest_call(item):
+    func_only = get_func_only_setting(item)
+    if func_only is True:
+        timeout_setup(item)
+    outcome = yield
+    if func_only is True:
+        timeout_teardown(item)
+
+
+@pytest.hookimpl(tryfirst=True)
+def pytest_report_header(config):
+    if config._env_timeout:
+        return ["timeout: %ss\ntimeout method: %s\ntimeout func_only: %s" %
+                (config._env_timeout,
+                 config._env_timeout_method,
+                 config._env_timeout_func_only)]
 
 
 @pytest.hookimpl(tryfirst=True)
@@ -92,21 +130,22 @@ SUPPRESS_TIMEOUT = False
 
 def timeout_setup(item):
     """Setup up a timeout trigger and handler"""
-    timeout, method = get_params(item)
-    if timeout is None or timeout <= 0:
+    params = get_params(item)
+    if params.timeout is None or params.timeout <= 0:
         return
-    if method == 'signal':
+    if params.method == 'signal':
         def handler(signum, frame):
             __tracebackhide__ = True
-            timeout_sigalrm(item, timeout)
+            timeout_sigalrm(item, params.timeout)
         def cancel():
-            signal.alarm(0)
+            signal.setitimer(signal.ITIMER_REAL, 0)
             signal.signal(signal.SIGALRM, signal.SIG_DFL)
         item.cancel_timeout = cancel
         signal.signal(signal.SIGALRM, handler)
-        signal.alarm(timeout)
-    elif method == 'thread':
-        timer = threading.Timer(timeout, timeout_timer, (item, timeout))
+        signal.setitimer(signal.ITIMER_REAL, params.timeout)
+    elif params.method == 'thread':
+        timer = threading.Timer(params.timeout, timeout_timer,
+                                (item, params.timeout))
         def cancel():
             timer.cancel()
             timer.join()
@@ -125,32 +164,65 @@ def timeout_teardown(item):
         cancel()
 
 
-def get_params(item):
-    """Return (timeout, method) for an item"""
-    timeout = method = None
-    if 'timeout' in item.keywords:
-        timeout, method = _parse_marker(item.keywords['timeout'])
-        timeout = _validate_timeout(timeout, 'marker')
-        method = _validate_method(method, 'marker')
-    if timeout is None:
-        timeout = item.config.getvalue('timeout')
+def get_env_settings(config):
+    timeout = config.getvalue('timeout')
     if timeout is None:
         timeout = _validate_timeout(
             os.environ.get('PYTEST_TIMEOUT'),
             'PYTEST_TIMEOUT environment variable')
     if timeout is None:
-        ini = item.config.getini('timeout')
+        ini = config.getini('timeout')
         if ini:
             timeout = _validate_timeout(ini, 'config file')
+
+    method = config.getvalue('timeout_method')
     if method is None:
-        method = item.config.getvalue('timeout_method')
-    if method is None:
-        ini = item.config.getini('timeout_method')
+        ini = config.getini('timeout_method')
         if ini:
             method = _validate_method(ini, 'config file')
     if method is None:
         method = DEFAULT_METHOD
-    return timeout, method
+
+    func_only = config.getini('timeout_func_only')
+    if func_only == []:
+        # No value set
+        func_only = None
+    if func_only is not None:
+        func_only = _validate_func_only(func_only, 'config file')
+    return Settings(timeout, method, func_only or False)
+
+
+def get_func_only_setting(item):
+    """Return the func_only setting for an item"""
+    func_only = None
+    marker = item.get_closest_marker('timeout')
+    if marker:
+        settings = get_params(item, marker=marker)
+        func_only = _validate_func_only(settings.func_only, 'marker')
+    if func_only is None:
+        func_only = item.config._env_timeout_func_only
+    if func_only is None:
+        func_only = False
+    return func_only
+
+
+def get_params(item, marker=None):
+    """Return (timeout, method) for an item"""
+    timeout = method = func_only = None
+    if not marker:
+        marker = item.get_closest_marker('timeout')
+    if marker is not None:
+        settings = _parse_marker(item.get_closest_marker(name='timeout'))
+        timeout = _validate_timeout(settings.timeout, 'marker')
+        method = _validate_method(settings.method, 'marker')
+        func_only = _validate_func_only(settings.func_only, 'marker')
+    if timeout is None:
+        timeout = item.config._env_timeout
+    if method is None:
+        method = item.config._env_timeout_method
+    if func_only is None:
+        func_only = item.config._env_timeout_func_only
+    return Settings(timeout, method, func_only)
 
 
 def _parse_marker(marker):
@@ -161,12 +233,14 @@ def _parse_marker(marker):
     """
     if not marker.args and not marker.kwargs:
         raise TypeError('Timeout marker must have at least one argument')
-    timeout = method = NOTSET = object()
+    timeout = method = func_only = NOTSET = object()
     for kw, val in marker.kwargs.items():
         if kw == 'timeout':
             timeout = val
         elif kw == 'method':
             method = val
+        elif kw == 'func_only':
+            func_only = val
         else:
             raise TypeError(
                 'Invalid keyword argument for timeout marker: %s' % kw)
@@ -186,7 +260,9 @@ def _parse_marker(marker):
         timeout = None
     if method is NOTSET:
         method = None
-    return timeout, method
+    if func_only is NOTSET:
+        func_only = None
+    return Settings(timeout, method, func_only)
 
 
 def _validate_timeout(timeout, where):
@@ -194,7 +270,7 @@ def _validate_timeout(timeout, where):
     if timeout is None:
         return None
     try:
-        return int(timeout)
+        return float(timeout)
     except ValueError:
         raise ValueError('Invalid timeout %s from %s' % (timeout, where))
 
@@ -208,6 +284,15 @@ def _validate_method(method, where):
     return method
 
 
+def _validate_func_only(func_only, where):
+    """Helper for get_params()"""
+    if func_only is None:
+        return False
+    if not isinstance(func_only, bool):
+        raise ValueError('Invalid func_only value %s from %s' % (func_only, where))
+    return func_only
+
+
 def timeout_sigalrm(item, timeout):
     """Dump stack of threads and raise an exception
 
@@ -238,7 +323,10 @@ def timeout_timer(item, timeout):
     try:
         capman = item.config.pluginmanager.getplugin('capturemanager')
         if capman:
-            stdout, stderr = capman.suspendcapture(item)
+            if pytest.__version__ >= '3.3':
+                stdout, stderr = capman.suspend_global_capture(item)
+            else:
+                stdout, stderr = capman.suspendcapture(item)
         else:
             stdout, stderr = None, None
         write_title('Timeout', sep='+')
